@@ -1,26 +1,43 @@
 %%%----------------------------------------------------------------------
 %%% File    : ejabberd_receiver.erl
-%%% Author  : Alexey Shchepin <alexey@sevcom.net>
+%%% Author  : Alexey Shchepin <alexey@process-one.net>
 %%% Purpose : Socket receiver for C2S and S2S connections
-%%% Created : 10 Nov 2003 by Alexey Shchepin <alexey@sevcom.net>
-%%% Id      : $Id: ejabberd_receiver.erl 537 2006-04-22 03:35:13Z alexey $
+%%% Created : 10 Nov 2003 by Alexey Shchepin <alexey@process-one.net>
+%%%
+%%%
+%%% ejabberd, Copyright (C) 2002-2008   Process-one
+%%%
+%%% This program is free software; you can redistribute it and/or
+%%% modify it under the terms of the GNU General Public License as
+%%% published by the Free Software Foundation; either version 2 of the
+%%% License, or (at your option) any later version.
+%%%
+%%% This program is distributed in the hope that it will be useful,
+%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
+%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+%%% General Public License for more details.
+%%%                         
+%%% You should have received a copy of the GNU General Public License
+%%% along with this program; if not, write to the Free Software
+%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+%%% 02111-1307 USA
+%%%
 %%%----------------------------------------------------------------------
 
 -module(ejabberd_receiver).
--author('alexey@sevcom.net').
--vsn('$Revision: 537 $ ').
+-author('alexey@process-one.net').
 
 -behaviour(gen_server).
 
 %% API
--export([start_link/5,
+-export([start_link/4,
 	 start/3,
 	 start/4,
 	 change_shaper/2,
 	 reset_stream/1,
 	 starttls/2,
 	 compress/2,
-	 become_controller/1,
+	 become_controller/2,
 	 close/1]).
 
 %% gen_server callbacks
@@ -44,9 +61,9 @@
 %% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
 %% Description: Starts the server
 %%--------------------------------------------------------------------
-start_link(Socket, SockMod, Shaper, MaxStanzaSize, C2SPid) ->
+start_link(Socket, SockMod, Shaper, MaxStanzaSize) ->
     gen_server:start_link(
-      ?MODULE, [Socket, SockMod, Shaper, MaxStanzaSize, C2SPid], []).
+      ?MODULE, [Socket, SockMod, Shaper, MaxStanzaSize], []).
 
 %%--------------------------------------------------------------------
 %% Function: start() -> {ok,Pid} | ignore | {error,Error}
@@ -58,7 +75,7 @@ start(Socket, SockMod, Shaper) ->
 start(Socket, SockMod, Shaper, MaxStanzaSize) ->
     {ok, Pid} = supervisor:start_child(
 		  ejabberd_receiver_sup,
-		  [Socket, SockMod, Shaper, MaxStanzaSize, self()]),
+		  [Socket, SockMod, Shaper, MaxStanzaSize]),
     Pid.
 
 change_shaper(Pid, Shaper) ->
@@ -73,8 +90,8 @@ starttls(Pid, TLSSocket) ->
 compress(Pid, ZlibSocket) ->
     gen_server:call(Pid, {compress, ZlibSocket}).
 
-become_controller(Pid) ->
-    gen_server:call(Pid, become_controller).
+become_controller(Pid, C2SPid) ->
+    gen_server:call(Pid, {become_controller, C2SPid}).
 
 close(Pid) ->
     gen_server:cast(Pid, close).
@@ -90,8 +107,7 @@ close(Pid) ->
 %%                         {stop, Reason}
 %% Description: Initiates the server
 %%--------------------------------------------------------------------
-init([Socket, SockMod, Shaper, MaxStanzaSize, C2SPid]) ->
-    XMLStreamState = xml_stream:new(C2SPid, MaxStanzaSize),
+init([Socket, SockMod, Shaper, MaxStanzaSize]) ->
     ShaperState = shaper:new(Shaper),
     Timeout = case SockMod of
 		  ssl ->
@@ -102,9 +118,7 @@ init([Socket, SockMod, Shaper, MaxStanzaSize, C2SPid]) ->
     {ok, #state{socket = Socket,
 		sock_mod = SockMod,
 		shaper_state = ShaperState,
-		c2s_pid = C2SPid,
 		max_stanza_size = MaxStanzaSize,
-		xml_stream_state = XMLStreamState,
 		timeout = Timeout}}.
 
 %%--------------------------------------------------------------------
@@ -120,7 +134,12 @@ handle_call({starttls, TLSSocket}, _From,
 	    #state{xml_stream_state = XMLStreamState,
 		   c2s_pid = C2SPid,
 		   max_stanza_size = MaxStanzaSize} = State) ->
-    xml_stream:close(XMLStreamState),
+    if
+	XMLStreamState /= undefined ->
+	    xml_stream:close(XMLStreamState);
+	true ->
+	    ok
+    end,
     NewXMLStreamState = xml_stream:new(C2SPid, MaxStanzaSize),
     NewState = State#state{socket = TLSSocket,
 			   sock_mod = tls,
@@ -154,10 +173,13 @@ handle_call(reset_stream, _From,
     NewXMLStreamState = xml_stream:new(C2SPid, MaxStanzaSize),
     Reply = ok,
     {reply, Reply, State#state{xml_stream_state = NewXMLStreamState}};
-handle_call(become_controller, _From, State) ->
-    activate_socket(State),
+handle_call({become_controller, C2SPid}, _From, State) ->
+    XMLStreamState = xml_stream:new(C2SPid, State#state.max_stanza_size),
+    NewState = State#state{c2s_pid = C2SPid,
+			   xml_stream_state = XMLStreamState},
+    activate_socket(NewState),
     Reply = ok,
-    {reply, Reply, State};
+    {reply, Reply, NewState};
 handle_call(_Request, _From, State) ->
     Reply = ok,
     {reply, Reply, State}.
@@ -231,7 +253,12 @@ handle_info(_Info, State) ->
 terminate(_Reason, #state{xml_stream_state = XMLStreamState,
 			  c2s_pid = C2SPid} = State) ->
     xml_stream:close(XMLStreamState),
-    gen_fsm:send_event(C2SPid, closed),
+    if
+	C2SPid /= undefined ->
+	    gen_fsm:send_event(C2SPid, closed);
+	true ->
+	    ok
+    end,
     catch (State#state.sock_mod):close(State#state.socket),
     ok.
 
@@ -248,11 +275,20 @@ code_change(_OldVsn, State, _Extra) ->
 
 activate_socket(#state{socket = Socket,
 		       sock_mod = SockMod}) ->
-    case SockMod of
-	gen_tcp ->
-	    inet:setopts(Socket, [{active, once}]);
-	_ ->
-	    SockMod:setopts(Socket, [{active, once}])
+    PeerName =
+	case SockMod of
+	    gen_tcp ->
+		inet:setopts(Socket, [{active, once}]),
+		inet:peername(Socket);
+	    _ ->
+		SockMod:setopts(Socket, [{active, once}]),
+		SockMod:peername(Socket)
+	end,
+    case PeerName of
+	{error, _Reason} ->
+	    self() ! {tcp_closed, Socket};
+	{ok, _} ->
+	    ok
     end.
 
 process_data(Data,
